"""Message module in charge of data serialization and deserialization based on the source or destination selector.

Message are automatically serialised based on the selector. The format follows these rules:

* Message are serialised using the protobuf format. Protobuf is used as it is a binary compact serialization format that
support bytes types with no intermediary (expansive) encoding.
* Message follow a hierarchy where the fields of a child selector must contain all the fields of the parent message. For
instance if message of selector /foo/bor has the definition:

{
color: str
size: int
}

The message with selector /foo/bar/baz must have the fields `color` and `size`, for instance:

{
color: str
size: int
weight: int
}

"""
import dataclasses
from typing import Any, Dict

from ostorlab.agent.message import proto_dict
from ostorlab.agent.message import serializer


@dataclasses.dataclass
class Message:
    """Message data class used for both incoming and outgoing messages."""

    selector: str
    data: Dict[str, Any]
    raw: bytes

    @classmethod
    def from_data(cls, selector: str, data: Dict[str, Any]) -> "Message":
        """Generate a message from a structured data and destination selector.

        This a convenience method to avoid directly handling protobuf messages and the not so friendly protobuf API.

        Args:
            selector: Target selector used to define the message format.
            data: Message data to be serialized to raw format.

        Returns:
            Message with both raw and data definitions.
        """
        raw = serializer.serialize(selector, data).SerializeToString()
        return cls(data=data, selector=selector, raw=raw)

    @classmethod
    def from_raw(cls, selector: str, raw: bytes) -> "Message":
        """Generate a message from a raw data and source selector.

        This a convenience method to avoid directly handling protobuf messages and the not so friendly protobuf API.

        Args:
            selector: Source selector used to define the message format.
            raw: Message raw data to be deserialized to pythonic data structure.

        Returns:
            Message with both raw and data definitions.
        """
        proto_message = serializer.deserialize(selector, raw)
        data = proto_dict.protobuf_to_dict(proto_message, use_enum_labels=True)
        return cls(data=data, selector=selector, raw=raw)



"""Runtime agent module gives access to methods to create the specific agent configurations & run the agent.

Usage
    agent_runtime = AgentRuntime(agent_settings, runtime_name, docker_client, bus_host,
                 bus_username, bus_password)
    agent_service = agent_runtime.create_agent_service(network_name, extra_configs)
"""
import hashlib
import io
import logging
import uuid
import base64
from typing import List, Optional

import docker
from docker import constants
from docker import errors
from docker.types import services as docker_types_services

from ostorlab import configuration_manager
from ostorlab import exceptions
from ostorlab.agent import definitions as agent_definitions
from ostorlab.runtimes import definitions

logger = logging.getLogger(__name__)

MOUNT_VARIABLES = {"$CONFIG_HOME": str(configuration_manager.OSTORLAB_PRIVATE_DIR)}

HEALTHCHECK_HOST = "0.0.0.0"
HEALTHCHECK_PORT = 5000
SECOND = 1000000000
HEALTHCHECK_RETRIES = 10
HEALTHCHECK_TIMEOUT = 30 * SECOND
HEALTHCHECK_START_PERIOD = 2 * SECOND
HEALTHCHECK_INTERVAL = 30 * SECOND


class Error(exceptions.OstorlabError):
    """Base Error."""


class MissingAgentDefinitionLabel(Error):
    """Agent definition label is missing from the agent image. This is likely due to the agent built directly with a
    docker command and not agent build."""


def _parse_mount_string_windows(string):
    """Handles parsing mounts on Windows OS."""
    parts = string.split(":")
    if len(parts) == 1:
        # This supposes only having a target like /root.
        return docker_types_services.Mount(target=parts[0], source=None)
    elif len(parts) == 2:
        # This supposes having a source linux style and target. Like /var/run/docker.sock:/var/run/docker.sock.
        target = parts[1]
        source = parts[0]
        mount_type = "volume"
        if source.startswith("/"):
            # Paths likes /var/run/docker.sock map to //var/run/docker.sock on windows.
            source = f"/{source}"
            mount_type = "bind"
        return docker_types_services.Mount(
            target, source, read_only=False, type=mount_type
        )
    elif len(parts) == 3:
        # This supposes two cases. First case is windows style with linux, like C:/Users/bob:/root. The second
        # case is /root:/root:ro
        if parts[2] in ("ro", "rw", "z", "Z"):
            target = parts[1]
            source = parts[0]
            mount_type = "volume"
            if source.startswith("/"):
                # Paths likes /var/run/docker.sock map to //var/run/docker.sock on windows.
                source = f"/{source}"
                mount_type = "bind"
            read_only = not parts[2] == "rw"
            return docker_types_services.Mount(
                target, source, read_only=read_only, type=mount_type
            )
        else:
            target = parts[2]
            source = ":".join(parts[:2])
            source = source.replace("\\", "/")
            mount_type = "bind"
            return docker_types_services.Mount(
                target, source, read_only=False, type=mount_type
            )
    elif len(parts) == 4:
        # This covers the case C:/Users/bob:/root:ro
        target = parts[2]
        source = ":".join(parts[:2])
        source = source.replace("\\", "/")
        mount_type = "bind"
        read_only = not (len(parts) == 3 or parts[3] == "rw")
        return docker_types_services.Mount(
            target, source, read_only=read_only, type=mount_type
        )
    else:
        raise errors.InvalidArgument(f'Invalid mount format "{string}"')


def _parse_mount_string_unix(string):
    """Handles parsing of mounts on Unix OS like Linux, Cygwin and MacOS."""
    parts = string.split(":")
    if len(parts) > 3:
        raise errors.InvalidArgument(f'Invalid mount format "{string}"')
    if len(parts) == 1:
        return docker_types_services.Mount(target=parts[0], source=None)
    else:
        target = parts[1]
        source = parts[0]
        mount_type = "volume"
        if source.startswith("/"):
            mount_type = "bind"
        read_only = not (len(parts) == 2 or parts[2] == "rw")
        return docker_types_services.Mount(
            target, source, read_only=read_only, type=mount_type
        )


def _parse_mount_string(string):
    """This is a fix to a bug in the Docker Python API by monkey patching the buggy method.

    The method is knowingly not supporting Windows. Until a fix PR is merged, we are hot patching the method to handle
    windows paths with : like C:/Users/bob/:/root.
    """
    if constants.IS_WINDOWS_PLATFORM:
        return _parse_mount_string_windows(string)
    else:
        return _parse_mount_string_unix(string)


docker_types_services.Mount.parse_mount_string = _parse_mount_string


class AgentRuntime:
    """Class to consolidate the agent settings and agent default definition, and create the agent service."""

    def __init__(
        self,
        agent_settings: definitions.AgentSettings,
        runtime_name: str,
        docker_client: docker.DockerClient,
        bus_url: str,
        bus_vhost: str,
        bus_management_url: str,
        bus_exchange_topic: str,
        redis_url: str,
        tracing_collector_url: str,
        gcp_logging_credential: Optional[str] = None,
    ) -> None:
        """Prepare all the necessary attributes for the agent runtime.

        Args:
            agent_settings: Agent settings object that will updated with bus configs.
            runtime_name: runtime name to add to all object names.
            docker_client: standard docker client.
            bus_url: Bus URL, may contain credentials.
            bus_vhost: Bus virtual host, common default is / but none is provided here.
            bus_management_url: Bus management URL, typically runs on a separate port over https.
            bus_exchange_topic: Bus exchange topic.
            redis_url: Redis URL.
            tracing_collector_url: Tracing Collector supporting Open Telemetry URL. The URL is a custom format to pass
             exporter and its arguments.
            gcp_logging_credential: GCP Logging JSON credentials.
        """
        self._uuid = uuid.uuid4()
        self._docker_client = docker_client
        self.agent = agent_settings
        self.image_name = agent_settings.container_image.split(":", maxsplit=1)[0]
        self.runtime_name = runtime_name
        self.bus_url = bus_url
        self.bus_vhost = bus_vhost
        self.bus_management_url = bus_management_url
        self.bus_exchange_topic = bus_exchange_topic
        self.redis_url = redis_url
        self.tracing_collector_url = tracing_collector_url
        self._gcp_logging_credential = gcp_logging_credential
        self.update_agent_settings()

    def create_settings_config(self) -> docker.types.ConfigReference:
        """Create a docker configuration of the  agent settings.

        Returns:
            docker ConfigReference of the settings configuration
        """
        agent_instance_settings_proto = self.agent.to_raw_proto()
        # The name is hashed to work around size limitation of Docker Config.
        config_name = hashlib.md5(
            f"config_settings_{self.image_name}_{self.runtime_name}_{self._uuid}".encode()
        ).hexdigest()

        try:
            settings_config = self._docker_client.configs.get(config_name)
            logging.warning(
                "found existing config %s, config will removed", config_name
            )
            settings_config.remove()
        except docker.errors.NotFound:
            logging.debug("all good, config %s is new", config_name)

        docker_config = self._docker_client.configs.create(
            name=config_name,
            labels={"ostorlab.universe": self.runtime_name},
            data=agent_instance_settings_proto,
        )
        return docker.types.ConfigReference(
            config_id=docker_config.id,
            config_name=config_name,
            filename="/tmp/settings.binproto",
        )

    def create_definition_config(self) -> docker.types.ConfigReference:
        """Create a docker configuration of the  agent definition.

        Returns:
            docker configuration reference of the agent defintion configuration.
        """
        agent_definition = self._docker_client.images.get(
            self.agent.container_image
        ).labels.get("agent_definition")
        # The name is hashed to work around size limitation of Docker Config.
        config_name = hashlib.md5(
            f"config_definition_{self.image_name}__{self.runtime_name}_{self._uuid}".encode()
        ).hexdigest()

        try:
            settings_config = self._docker_client.configs.get(config_name)
            logging.warning(
                "found existing config %s, config will removed", config_name
            )
            settings_config.remove()
        except docker.errors.NotFound:
            logging.debug("all good, config %s is new", config_name)

        docker_config = self._docker_client.configs.create(
            name=config_name,
            labels={"ostorlab.universe": self.runtime_name},
            data=str.encode(agent_definition),
        )
        return docker.types.ConfigReference(
            config_id=docker_config.id,
            config_name=config_name,
            filename="/tmp/ostorlab.yaml",
        )

    def create_agent_definition_from_label(self) -> agent_definitions.AgentDefinition:
        """Read the agent yaml definition from the docker image labels.

        Returns:
            the agent definition.
        """
        docker_image = self._docker_client.images.get(self.agent.container_image)
        yaml_definition_string = docker_image.labels.get("agent_definition")
        if yaml_definition_string is None:
            raise MissingAgentDefinitionLabel(
                f"agent definition label is missing from image {docker_image.tags[0]}"
            )
        with io.StringIO(yaml_definition_string) as file:
            agent_definition = agent_definitions.AgentDefinition.from_yaml(file)
            return agent_definition

    def update_agent_settings(self) -> None:
        """Update agent settings with values from the local runtime."""
        self.agent.bus_url = self.bus_url
        self.agent.bus_exchange_topic = self.bus_exchange_topic
        self.agent.bus_management_url = self.bus_management_url
        self.agent.bus_vhost = self.bus_vhost
        self.agent.healthcheck_host = HEALTHCHECK_HOST
        self.agent.healthcheck_port = HEALTHCHECK_PORT
        self.agent.redis_url = self.redis_url
        self.agent.tracing_collector_url = self.tracing_collector_url

    def create_docker_healthchek(self) -> docker.types.Healthcheck:
        """Create a docker healthcheck configuration for the agent service.

        Returns:
            docker healthcheck configuration.
        """
        # wait 2s and check max 5 times with 0.5s between each check.
        healthcheck = docker.types.Healthcheck(
            test=["CMD", "ostorlab", "agent", "healthcheck"],
            retries=HEALTHCHECK_RETRIES,
            timeout=HEALTHCHECK_TIMEOUT,
            start_period=HEALTHCHECK_START_PERIOD,
            interval=HEALTHCHECK_INTERVAL,
        )
        return healthcheck

    def replace_variable_mounts(self, mounts: List[str]):
        """Replace path variables for the container mounts

        Args:
            mounts: List of src:dst paths to mount
        """

        replaced_mounts = []
        for mount in mounts:
            for mount_variable, mount_value in MOUNT_VARIABLES.items():
                mount = mount.replace(mount_variable, mount_value)
            replaced_mounts.append(mount)
        return replaced_mounts

    def create_agent_service(
        self,
        network_name: str,
        extra_configs: Optional[List[docker.types.ConfigReference]] = None,
        extra_mounts: Optional[List[docker.types.Mount]] = None,
    ) -> docker.models.services.Service:
        """Create the docker agent service with proper configs and policies.

        Args:
            network_name: network name to attach the service to.
            extra_configs: list of docker ConfigReferences that will be exposed to the service.

        Returns:
            the agent docker service.
        """
        agent_definition = self.create_agent_definition_from_label()
        self.agent.open_ports = self.agent.open_ports or agent_definition.open_ports
        if self.agent.open_ports:
            endpoint_spec = docker_types_services.EndpointSpec(
                ports={p.destination_port: p.source_port for p in self.agent.open_ports}
            )
        else:
            endpoint_spec = docker_types_services.EndpointSpec(mode="dnsrr")

        configs = []
        configs.append(self.create_settings_config())
        configs.append(self.create_definition_config())
        if extra_configs is not None:
            configs.extend(extra_configs)

        mounts = self.agent.mounts or agent_definition.mounts
        mounts = self.replace_variable_mounts(mounts)
        if extra_mounts is not None:
            mounts.extend(extra_mounts)

        constraints = self.agent.constraints or agent_definition.constraints
        mem_limit = self.agent.mem_limit or agent_definition.mem_limit
        restart_policy = (
            self.agent.restart_policy or agent_definition.restart_policy or "any"
        )
        caps = self.agent.caps or agent_definition.caps

        service_name = (
            agent_definition.service_name
            or self.agent.container_image.replace(":", "_").replace(".", "")
            + "_"
            + self.runtime_name
        )

        env = [
            f"UNIVERSE={self.runtime_name}",
        ]
        if self._gcp_logging_credential is not None:
            env.append(
                f"GCP_LOGGING_CREDENTIAL={base64.b64encode(self._gcp_logging_credential.encode()).decode()}"
            )

        agent_service = self._docker_client.services.create(
            image=self.agent.container_image,
            networks=[network_name],
            env=env,
            name=service_name,
            restart_policy=docker_types_services.RestartPolicy(
                condition=restart_policy
            ),
            mounts=mounts,
            healthcheck=self.create_docker_healthchek(),
            labels={"ostorlab.universe": self.runtime_name},
            configs=configs,
            constraints=constraints,
            endpoint_spec=endpoint_spec,
            resources=docker_types_services.Resources(mem_limit=mem_limit),
            cap_add=caps,
        )

        return agent_service



"""Vulnz List command."""
import logging

import click

from ostorlab.cli import console as cli_console
from ostorlab.cli.vulnz import vulnz

console = cli_console.Console()

logger = logging.getLogger(__name__)


@vulnz.command(name="list")
@click.option("--scan-id", "-s", "scan_id", help="Id of the scan.", required=True)
@click.pass_context
def list_cli(ctx: click.core.Context, scan_id: int) -> None:
    """CLI command to list vulnerabilities for a scan."""

    runtime_instance = ctx.obj["runtime"]
    console.info(f"Fetching vulnerabilities for scan {scan_id}")
    runtime_instance.list_vulnz(scan_id=scan_id)



"""ci_scan module that handles running a scan on the CI."""
from ostorlab.cli.rootcli import rootcli


@rootcli.group()
def ci_scan() -> None:
    """You can use ci_scan to run scan on the CI.\n"""
    pass



"""Agent search command."""
import logging
from typing import Dict

import click
from rich import markdown

from ostorlab import configuration_manager
from ostorlab.apis import agent_search as agent_search_api
from ostorlab.apis.runners import public_runner, authenticated_runner
from ostorlab.apis.runners import runner as base_runner
from ostorlab.cli import console as cli_console
from ostorlab.cli.agent import agent

console = cli_console.Console()

logger = logging.getLogger(__name__)


@agent.command(name="search")
@click.option("--keyword", "-k", help="Keyword to use for the search.", required=True)
def search_cli(keyword: str) -> None:
    """CLI command to search for agents in the Agents store."""
    result_agents = search_agents(keyword).get("agents")
    console.success("Search agents done.")
    agents = []
    version = ""
    description = ""
    in_selectors = ""
    out_selectors = ""

    if result_agents is not None:
        for result_agent in result_agents:
            versions = result_agent["versions"].get("versions")
            if len(versions) > 0:
                version = versions[0].get("version")
                description = versions[0].get("description")
                in_selectors = ", ".join(versions[0].get("inSelectors"))
                out_selectors = ",".join(versions[0].get("outSelectors"))
            else:
                version = "-"
                description = "-"
                in_selectors = "-"
                out_selectors = "-"

            agents.append(
                {
                    "key": result_agent["key"],
                    "version": version,
                    "description": markdown.Markdown(description),
                    "in_selectors": in_selectors,
                    "out_selectors": out_selectors,
                }
            )

    columns = {
        "Agent key": "key",
        "Version": "version",
        "Description": "description",
        "In selectors": "in_selectors",
        "Out selectors": "out_selectors",
    }
    title = f"Found {len(agents)} Agents"
    console.table(columns=columns, data=agents, title=title)


def search_agents(search: str) -> Dict:
    """Sends an API request with the agent key, and retrieve the agent information.

    Args:
        search: the keyword to search for

    Returns:
        List of agent matching the search.

    Raises:
        click Exit exception with satus code 2 when API response is invalid.
    """
    config_manager = configuration_manager.ConfigurationManager()

    if config_manager.is_authenticated:
        runner = authenticated_runner.AuthenticatedAPIRunner()
    else:
        runner = public_runner.PublicAPIRunner()

    try:
        response = runner.execute(agent_search_api.AgentSearchAPIRequest(search))
    except base_runner.ResponseError as e:
        console.error("Requested resource not found.")
        raise click.exceptions.Exit(2) from e

    if "errors" in response:
        error_message = (
            f"The provided search : {search} does not correspond to any agent."
        )
        console.error(error_message)
        raise click.exceptions.Exit(2)
    else:
        agent_details = response["data"]["agents"]
        return agent_details



"""Module for the root command: vulnz."""
from ostorlab.cli.vulnz.vulnz import vulnz
from ostorlab.cli.vulnz import list as list_cli
from ostorlab.cli.vulnz import describe as describe_cli
from ostorlab.cli.vulnz import dump



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: ostorlab/scanner/proto/assets/link.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n(ostorlab/scanner/proto/assets/link.proto\x12\"ostorlab.scanner.proto.assets.link\"%\n\x06Header\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t\"%\n\x06\x43ookie\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t\"7\n\x0cLocalStorage\x12\x0b\n\x03url\x18\x01 \x01(\t\x12\x0b\n\x03key\x18\x02 \x01(\t\x12\r\n\x05value\x18\x03 \x01(\t\"6\n\tFormInput\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t\x12\x0c\n\x04type\x18\x03 \x01(\t\"V\n\x04\x46orm\x12=\n\x06inputs\x18\x01 \x03(\x0b\x32-.ostorlab.scanner.proto.assets.link.FormInput\x12\x0f\n\x07\x65nctype\x18\x02 \x01(\t\">\n\x0e\x46ormCredential\x12\r\n\x05login\x18\x01 \x01(\t\x12\x10\n\x08password\x18\x02 \x01(\t\x12\x0b\n\x03url\x18\x03 \x01(\t\"\"\n\x10ScriptCredential\x12\x0e\n\x06script\x18\x01 \x01(\t\"\xbf\x04\n\x07Message\x12\x0b\n\x03url\x18\x01 \x01(\t\x12\x0e\n\x06method\x18\x02 \x01(\t\x12\x0e\n\x04\x62ody\x18\x03 \x01(\x0cH\x00\x12\x38\n\x04\x66orm\x18\x04 \x01(\x0b\x32(.ostorlab.scanner.proto.assets.link.FormH\x00\x12\x41\n\rextra_headers\x18\x05 \x03(\x0b\x32*.ostorlab.scanner.proto.assets.link.Header\x12;\n\x07\x63ookies\x18\x06 \x03(\x0b\x32*.ostorlab.scanner.proto.assets.link.Cookie\x12M\n\x0f\x66orm_credential\x18\x07 \x01(\x0b\x32\x32.ostorlab.scanner.proto.assets.link.FormCredentialH\x01\x12Q\n\x11script_credential\x18\x08 \x01(\x0b\x32\x34.ostorlab.scanner.proto.assets.link.ScriptCredentialH\x01\x12;\n\x06parent\x18\t \x01(\x0b\x32+.ostorlab.scanner.proto.assets.link.Message\x12L\n\x12localstorage_items\x18\n \x03(\x0b\x32\x30.ostorlab.scanner.proto.assets.link.LocalStorageB\x0c\n\nbody_oneofB\x12\n\x10\x63redential_oneof')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'ostorlab.scanner.proto.assets.link_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _HEADER._serialized_start=80
  _HEADER._serialized_end=117
  _COOKIE._serialized_start=119
  _COOKIE._serialized_end=156
  _LOCALSTORAGE._serialized_start=158
  _LOCALSTORAGE._serialized_end=213
  _FORMINPUT._serialized_start=215
  _FORMINPUT._serialized_end=269
  _FORM._serialized_start=271
  _FORM._serialized_end=357
  _FORMCREDENTIAL._serialized_start=359
  _FORMCREDENTIAL._serialized_end=421
  _SCRIPTCREDENTIAL._serialized_start=423
  _SCRIPTCREDENTIAL._serialized_end=457
  _MESSAGE._serialized_start=460
  _MESSAGE._serialized_end=1035
# @@protoc_insertion_point(module_scope)



"""Define methods to style components for console."""


def style_risk(risk: str) -> str:
    """Stylize the risk with colors."""
    if risk == "HIGH":
        return "[bold bright_white on red]High[/]"
    elif risk == "MEDIUM":
        return "[bold bright_white on yellow]Medium[/]"
    elif risk == "LOW":
        return "[bold bright_white on bright_yellow]Low[/]"
    else:
        return risk


def style_progress(progress: str) -> str:
    """Stylize the scan progress with colors."""
    if progress == "done":
        return "[bold green4]Done[/]"
    if progress == "error":
        return "[bold magenta]Error[/]"
    if progress == "not_started":
        return "[bold bright_black]Not Started[/]"
    if progress == "stopped":
        return "[bold bright_red]Stopped[/]"
    if progress == "in_progress":
        return "[bold bright_cyan]Running[/]"
    else:
        return progress


def style_asset(asset: str) -> str:
    """Stylize the scan asset with colors and emojis."""
    if asset == "android_store":
        return "[bold green4]:iphone: Android Store[/]"
    elif asset == "ios_store":
        return "[bold bright_white]:apple: iOS Store[/]"
    elif asset == "android":
        return "[bold bright_green]:iphone: Android[/]"
    elif asset == "ios":
        return "[bold white]:apple: iOS[/]"
    else:
        return asset



"""Defines call back to trigger a scan after receiving a startAgentScan messages in the NATS."""
import logging
import ipaddress
from typing import List, Any, Optional

import docker

from ostorlab.runtimes import registry
from ostorlab.runtimes import definitions
from ostorlab.runtimes import runtime
from ostorlab.cli import install_agent
from ostorlab.assets import asset
from ostorlab.assets import ipv4
from ostorlab.assets import ipv6
from ostorlab.assets import android_store
from ostorlab.assets import android_aab
from ostorlab.assets import android_apk
from ostorlab.assets import file
from ostorlab.assets import ios_ipa
from ostorlab.assets import domain_name
from ostorlab.assets import link as link_asset
from ostorlab.assets import ios_store
from ostorlab.assets import agent as agent_asset
from ostorlab.utils import scanner_state_reporter
from ostorlab.scanner import scanner_conf


logger = logging.getLogger(__name__)


def _install_agents(
    runtime_instance: runtime.Runtime,
    agents,
    docker_client: Optional[docker.DockerClient] = None,
) -> None:
    """Trigger installation of the agents that will run the scan."""
    try:
        runtime_instance.install(docker_client=docker_client)
        for agent in agents:
            install_agent.install(
                agent_key=agent.key, version=agent.version, docker_client=docker_client
            )
    except install_agent.AgentDetailsNotFound:
        logger.warning("agent %s not found on the store", agent.key)


def _prepare_ip_asset(ip_request) -> asset.Asset:
    """Return IP assets from a NATs received message."""
    ip_network = ipaddress.ip_network(ip_request.host, strict=False)
    if ip_network.version == 4:
        return ipv4.IPv4(
            host=ip_network.network_address.exploded,
            mask=ip_request.mask or str(ip_network.prefixlen),
        )
    elif ip_network.version == 6:
        return ipv6.IPv6(
            host=ip_network.network_address.exploded,
            mask=ip_request.mask or str(ip_network.prefixlen),
        )
    else:
        raise ValueError(f"Invalid Ip address {ip_request.host}")


def _extract_assets(request: Any) -> List[asset.Asset]:
    """Returns list of specific Ostorlab-injectable assets, from a message received from NATs."""
    logger.debug("Extracting assets.")
    assets = []
    asset_type = request.WhichOneof("asset")
    if asset_type in ("ip", "ip4v", "ipv6"):
        ip_request = request.ip or request.ipv4 or request.ipv6
        assets.append(_prepare_ip_asset(ip_request))

    elif asset_type == "android_store":
        assets.append(
            android_store.AndroidStore(package_name=request.android_store.package_name)
        )

    elif asset_type == "ios_store":
        assets.append(ios_store.IOSStore(bundle_id=request.ios_store.bundle_id))

    elif asset_type == "ipa":
        assets.append(
            ios_ipa.IOSIpa(
                content=request.ipa.content,
                path=request.ipa.path,
                content_url=request.ipa.content_url,
            )
        )
    elif asset_type == "apk":
        assets.append(
            android_apk.AndroidApk(
                content=request.apk.content,
                path=request.apk.path,
                content_url=request.apk.content_url,
            )
        )
    elif asset_type == "aab":
        assets.append(
            android_aab.AndroidAab(
                content=request.aab.content,
                path=request.aab.path,
                content_url=request.aab.content_url,
            )
        )
    elif asset_type == "domain":
        assets.append(domain_name.DomainName(name=request.domain_name.name))

    elif asset_type == "agent":
        assets.append(
            agent_asset.Agent(
                key=request.agent.key,
                version=request.agent.version,
                git_location=request.agent.git_location,
                docker_location=request.agent.docker_location,
                yaml_file_location=request.agent.yaml_file_location,
            )
        )

    elif asset_type == "file":
        assets.append(
            file.File(
                content=request.file.content,
                path=request.file.path,
                content_url=request.file.content_url,
            )
        )
    elif asset_type == "network":
        for ip in request.network.ips:
            ip_asset = _prepare_ip_asset(ip)
            assets.append(ip_asset)

    elif asset_type == "links":
        for link in request.links.links:
            assets.append(
                link_asset.Link(
                    url=link.url,
                    method=link.url or "GET",
                )
            )

    else:
        logger.error(
            "%s not supported from scan reference  %s ",
            asset_type,
            request.reference_scan_id,
        )

    logger.debug("Extracted assets: %s.", assets)
    return assets


def _extract_agent_group_definition(request: Any) -> definitions.AgentGroupDefinition:
    agent_group_definition = definitions.AgentGroupDefinition.from_bus_message(request)
    logger.debug("Extracted agent group definition: %s.", agent_group_definition)
    return agent_group_definition


def _extract_scan_id(request: Any) -> int:
    scan_id = int(request.scan_id)
    logger.debug("Extracted scan id: %s.", scan_id)
    return scan_id


def _update_state_reporter(
    state_reporter: scanner_state_reporter.ScannerStateReporter, scan_id: int
) -> scanner_state_reporter.ScannerStateReporter:
    state_reporter.scan_id = scan_id
    return state_reporter


def _connect_containers_registry(
    configuration: scanner_conf.RegistryConfig,
) -> docker.DockerClient:
    """Connect to container registry."""
    logger.debug("Connecting to private container registry.")
    client = docker.from_env()
    client.login(
        username=configuration.username,
        password=configuration.token,
        registry=configuration.url,
    )
    return client


def start_scan(
    subject: str,
    request: Any,
    state_reporter: scanner_state_reporter.ScannerStateReporter,
    registry_conf: scanner_conf.RegistryConfig,
) -> str:
    """Responsible for triggering an Ostorlab scan, after receiving a startAgentScan message in NATs.

    Args:
        subject: Subject of the received message.
        request: Deserialized message.
        state_reporter: State reporter instance responsible for sending current state of the scanner.
        registry_conf: Credentials to the registry, useful to pull agents images.
    """
    logger.debug("Triggering scan after receiving message on: %s", subject)
    docker_client = _connect_containers_registry(configuration=registry_conf)

    agent_group_definition = _extract_agent_group_definition(request)
    assets = _extract_assets(request)
    scan_id = _extract_scan_id(request)

    state_reporter = _update_state_reporter(state_reporter, scan_id)

    runtime_instance = registry.select_runtime(
        runtime_type="local", scan_id=str(scan_id), run_default_agents=False
    )

    if runtime_instance.can_run(agent_group_definition=agent_group_definition) is True:
        _install_agents(
            runtime_instance=runtime_instance,
            agents=request.agents,
            docker_client=docker_client,
        )

        runtime_instance.scan(
            agent_group_definition=agent_group_definition,
            assets=assets,
            title=None,
        )
        return runtime_instance.name
    else:
        logger.error(
            "The runtime does not support the provided agent list or group definition."
        )



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: ostorlab/scanner/proto/assets/apk.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\'ostorlab/scanner/proto/assets/apk.proto\x12!ostorlab.scanner.proto.assets.apk\"\'\n\x0f\x41ndroidMetadata\x12\x14\n\x0cpackage_name\x18\x01 \x01(\t\" \n\x0bIOSMetadata\x12\x11\n\tbundle_id\x18\x01 \x01(\t\"\xe7\x01\n\x07Message\x12\x0f\n\x07\x63ontent\x18\x01 \x01(\x0c\x12\x0c\n\x04path\x18\x02 \x01(\t\x12\x13\n\x0b\x63ontent_url\x18\x03 \x01(\t\x12N\n\x10\x61ndroid_metadata\x18\x04 \x01(\x0b\x32\x32.ostorlab.scanner.proto.assets.apk.AndroidMetadataH\x00\x12\x46\n\x0cios_metadata\x18\x05 \x01(\x0b\x32..ostorlab.scanner.proto.assets.apk.IOSMetadataH\x00\x42\x10\n\x0emetadata_oneof')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'ostorlab.scanner.proto.assets.apk_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _ANDROIDMETADATA._serialized_start=78
  _ANDROIDMETADATA._serialized_end=117
  _IOSMETADATA._serialized_start=119
  _IOSMETADATA._serialized_end=151
  _MESSAGE._serialized_start=154
  _MESSAGE._serialized_end=385
# @@protoc_insertion_point(module_scope)



"""KB provides automatic mapping from the KB entry folder name to the meta content."""
import dataclasses
import json
import pathlib

from typing import Dict

KB_FOLDER = "KB"
META_JSON = "meta.json"
DESCRIPTION = "description.md"
RECOMMENDATION = "recommendation.md"


@dataclasses.dataclass
class Entry:
    """KB entry with information like title, risk rating and description."""

    title: str
    risk_rating: str
    references: Dict[str, str]
    short_description: str = ""
    description: str = ""
    recommendation: str = ""
    security_issue: bool = False
    privacy_issue: bool = False
    has_public_exploit: bool = False
    targeted_by_malware: bool = False
    targeted_by_ransomware: bool = False
    targeted_by_nation_state: bool = False
    cvss_v3_vector: str = ""


class MetaKB(type):
    """Handles KB mapping to appropriate folder."""

    def __getattr__(cls, item: str) -> Entry:
        kb_path = pathlib.Path(__file__).parent / KB_FOLDER
        paths = [f for f in kb_path.glob(f"**/{item}") if f.is_dir()]
        if not paths:
            raise ValueError(f"{item} does not exists.")
        entry_path = paths[0]
        if not (entry_path / META_JSON).exists():
            raise ValueError(f"{entry_path} does not have a mapping.")
        with (entry_path / META_JSON).open(encoding="utf-8") as f, (
            entry_path / DESCRIPTION
        ).open(encoding="utf-8") as d, (entry_path / RECOMMENDATION).open(
            encoding="utf-8"
        ) as r:
            meta = json.loads(f.read())
            return Entry(
                title=meta.get("title"),
                risk_rating=meta.get("risk_rating"),
                short_description=meta.get("short_description"),
                description=d.read(),
                recommendation=r.read(),
                references=meta.get("references"),
                security_issue=meta.get("security_issue", False),
                privacy_issue=meta.get("privacy_issue", False),
                has_public_exploit=meta.get("has_public_exploit", False),
                targeted_by_malware=meta.get("targeted_by_malware", False),
                targeted_by_ransomware=meta.get("targeted_by_ransomware", False),
                targeted_by_nation_state=meta.get("targeted_by_nation_state", False),
                cvss_v3_vector=meta.get("cvss_v3_vector", ""),
            )


class KB(object, metaclass=MetaKB):
    """Vulnerability knowledge base dispatcher."""



"""This module triggers a scan using the android_store asset."""
import logging
from typing import Tuple, Optional

import click

from ostorlab.cli.scan.run import run
from ostorlab.cli import console as cli_console
from ostorlab.assets import android_store as android_store_asset

console = cli_console.Console()
logger = logging.getLogger(__name__)


@run.run.command(name="android-store")
@click.option("--package-name", multiple=True, required=False)
@click.pass_context
def android_store(
    ctx: click.core.Context, package_name: Optional[Tuple[str]] = ()
) -> None:
    """Run scan for a package_name."""
    runtime = ctx.obj["runtime"]
    assets = []
    if package_name != ():
        for package in package_name:
            assets.append(android_store_asset.AndroidStore(package_name=package))
    else:
        console.error("Command missing a package name.")
        raise click.exceptions.Exit(2)

    logger.debug("scanning assets %s", [str(asset) for asset in assets])
    runtime.scan(
        title=ctx.obj["title"],
        agent_group_definition=ctx.obj["agent_group_definition"],
        assets=assets,
    )



"""Representations of nats configuration definitions."""

import dataclasses
from typing import List, Dict, Any


@dataclasses.dataclass
class RegistryConfig:
    """Represents the configuration for container registry credentials."""

    username: str
    token: str
    url: str


@dataclasses.dataclass
class SubjectBusConfigs:
    """Represents the configuration for a subject and its corresponding queue."""

    subject: str
    queue: str


@dataclasses.dataclass
class ScannerConfig:
    """Represents the configuration for a scanner."""

    bus_url: str
    bus_cluster_id: str
    bus_client_name: str
    registry_conf: RegistryConfig
    subject_bus_configs: List[SubjectBusConfigs]

    @classmethod
    def from_json(cls, config: Dict[str, Any]):
        """Creates a ScannerConfig instance from a JSON configuration.

        Args:
            config: The JSON configuration.

        Returns:
            ScannerConfig: An instance of ScannerConfig.
        """

        subject_configs = config.get("data", {}).get("scanners", {}).get("scanners", [])

        if len(subject_configs) == 0:
            return None

        subject_configs = subject_configs[0].get("config", {})
        bus_configs = []
        for subject_config in subject_configs.get("subjectBusConfigs", {}).get(
            "subjectBusConfigs", []
        ):
            bus_configs.append(
                SubjectBusConfigs(
                    subject=subject_config.get("subject"),
                    queue=subject_config.get("queue"),
                )
            )
        conf = (
            config.get("data", {})
            .get("scanners", {})
            .get("scanners", [])[0]
            .get("config", {})
        )
        registry_conf = conf.get("registryConfiguration", {})
        registry_conf_instance = RegistryConfig(
            username=registry_conf.get("accountName"),
            token=registry_conf.get("credentials"),
            url=registry_conf.get("url"),
        )
        return cls(
            bus_url=conf.get("busUrl"),
            bus_cluster_id=conf.get("busClusterId"),
            bus_client_name=conf.get("busClientName"),
            registry_conf=registry_conf_instance,
            subject_bus_configs=bus_configs,
        )
